/*
 * This file is part of libslab.
 *
 * Copyright (c) 2006 Novell, Inc.
 *
 * Libslab is free software; you can redistribute it and/or modify it under the
 * terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * Libslab is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libslab; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


#ifndef __SLAB_H__
#define __SLAB_H__

#include <libslab/app-resizer.h>
#include <libslab/app-shell.h>
#include <libslab/application-tile.h>
#include <libslab/bookmark-agent.h>
#include <libslab/directory-tile.h>
#include <libslab/document-tile.h>
#include <libslab/double-click-detector.h>
#include <libslab/gnome-utils.h>
#include <libslab/libslab-utils.h>
#include <libslab/nameplate-tile.h>
#include <libslab/search-bar.h>
#include <libslab/search-context-picker.h>
#include <libslab/search-entry.h>
#include <libslab/shell-window.h>
#include <libslab/slab-gnome-util.h>
#include <libslab/slab-section.h>
#include <libslab/system-tile.h>
#include <libslab/tile.h>

#endif /* __SLAB_H__ */

